#!/usr/bin/env python3

"""
NAME:
	updsoa.py - update DNS SOA record

SYNOPSIS:
	updsoa.py [options] "soa_file" ...

DESCRIPTION:
	This script updates the serial number in each "soa_file".
	The serial number is formatted as "yyyymmddnn" where "nn"
	differentiates between multiple updates on the same day.
	For zones with serious flux, the option '-u' uses UTC as the serial
	number.

	Note: that UTC is less than yyyymmddnn so you cannot easily
        switch from one format to the other.

	The previous version of "soa_file" is left in "soa_file".old

	If "soa_file" does not exist, we create it.

	We need some configuration for that.
	Most can be defaulted, but we require a 'nameserver' ('ns')
	and 'ns_list', and will raise an exception if either are
	missing.

	An example configuration looks like::

		ns = ns.crufty.net
		hostmaster = hostmaster.crufty.net
		ns_list = ns.list
		# this is equivalent to the above
		zone_ns_list {=} default = ns.list
		zone_ns_list {=} external/crufty.net=external/ns.list
        	zone_ns_list {=} hosts/crufty.net = ns.crufty.net, ns2.crufty.net
		zone_ttl {=} default = 14400


	In the above, we have separate 'ns_list' for an internal and
	external view of the same domain, we also have 'default'
	entries that will be used by other zones.
	The '{=}' syntaxt indicates that the value is a key value pair
	that goes in a dictionary named before '{=}'.

	In addition to 'zone_ns_list' and 'zone_ttl' you can configure
	'zone_refresh', 'zone_retry', 'zone_expire', 'zone_min_ttl'
	and even 'zone_ns'.  All the non 'ns' fields have useful
	defaults.

	Options:

	-c "config"
		Read settings from "config", see conf.py for syntax.
		We will automatically read 'dns.cf' and '${progname}.cf'
		from the directory this script is in ('${progdir}'),
		as well as '.dnc.cf' and '.${progname}.cf' if they
		exist in the current directory or a parent.

		In fact unlike earlier iterations of this tool, we
		expect most settings to be provided via a "config" file.
		
	-o "var=value"

		This allows overriding random settings from "config".
        
	-u	Use current time in seconds as serial number.

	To cater for magic domains (eg. round-robin zones) we check
	for each of the SOA fields can be set per zone.
	
AUTHOR:
	Simon J. Gerraty <sjg@crufty.net>
"""

"""
RCSid:
	$Id: updsoa.py,v 1.6 2023/06/10 17:06:44 sjg Exp $

	@(#) Copyright (c) 1997-2023 Simon J. Gerraty.

	This file is provided in the hope that it will
	be of use.  There is absolutely NO WARRANTY.
	Permission to copy, redistribute or otherwise
	use this file is hereby granted provided that 
	the above copyright notice and this notice are
	left intact. 
      
	Please send copies of changes and bug-fixes to:
	sjg@crufty.net
"""


import sys
import os
import os.path
import time
import dnsutils
import conf as cf

def soa_field(zone, field, default, conf):
    """lookup field for zone

    create an empty dictionary in conf for field
    if needed, and use cf.lookup to find an entry for
    zone or 'default', if nothing found return default.
    """
    if not field in conf:
        conf[field] = {}
    return cf.lookup(conf[field], [zone, 'default'], default)

def soa_value(zone, field, default, conf):
    """return value from soa_field() as an int"""
    return int(soa_field(zone, field, default, conf))

def new_soa(soa, conf):
    """Generate a new SOA file

    We need ns and hostmaster for the SOA record,
    as well as ns list which we get from
    zone_ns_list[zone], zone_ns_list['default'] or
    ns_list.

    zone in the above lookup includes the directory we store the data
    in since hosts/zone and external/zone may need different name
    servers.

    Once we have ns_list it may identify a file or a list of name
    servers.
    """
    file = open(soa, 'w')
    zone = soa.strip('.soa')
    zfile = soa.replace('.soa', '.db')
    ns_list = soa_field(zone, 'zone_ns_list', conf.get('ns_list'), conf)
    ns = soa_field(zone, 'zone_ns', conf.get('ns'), conf)

    # ns and ns_list are required
    if not ns_list and not ns:
        raise ValueError('{}: no ns or ns_list'.format(soa))
    elif not ns:
        raise ValueError('{}: no ns'.format(soa))
    elif not ns_list:
        raise ValueError('{}: no ns_list'.format(soa))

    now = int(time.time())
    if 'serial_utc' in conf:
        serial = now
    else:
        serial = time.strftime('%Y%m%d00', time.localtime(now))

    ttl = soa_value(zone, 'zone_ttl', 14400, conf)
    refresh = soa_value(zone, 'zone_refresh', 14400, conf)
    retry = soa_value(zone, 'zone_retry', 3600, conf)
    expire = soa_value(zone, 'zone_expire', 2400000, conf)
    min_ttl = soa_value(zone, 'zone_min_ttl', 43200, conf)

    hostmaster = soa_field(zone, 'zone_hostmaster', conf.get('hostmaster', 'hostmaster.'+ns), conf)
    # this will get USER from env if needed
    user = cf.subst('${USER:Uhostmaster}', conf)
                           
    if not isinstance(ns_list, list):
        ns_list = ns_list.split(',')
    NS_LIST = None
    for n in ns_list:
        if n.endswith('.list'):
            NS_LIST = n
            break
        
    print(''';; DO NOT EDIT THIS FILE it is maintained by magic
$TTL\t{}
@\tIN\tSOA\t{}. {}. (
\t\t\t{:>10}\t; Last changed by - {}
\t\t\t{:10d}\t; Refresh
\t\t\t{:10d}\t; Retry
\t\t\t{:10d}\t; Expire
\t\t\t{:10d} )\t; Minimum TTL
'''.format(ttl,ns,hostmaster,serial,user,refresh,retry,expire,min_ttl), file=file)

    if NS_LIST:
        print('$INCLUDE {}'.format(NS_LIST), file=file)
    else:
        a = '@'
        for n in ns_list:
            print('{}\tIN\tNS\t{}.'.format(a,n), file=file)
            a = ''
        print('', file=file)
    print('$INCLUDE {}'.format(zfile), file=file)
    file.close()

def update_soa(soa, conf):
    if not os.path.exists(soa):
        return new_soa(soa, conf)

    soa_new = soa+'.new'
    oldfile = open(soa, 'r')
    newfile = open(soa_new, 'w')
    user = conf.get('USER', 'hostmaster')
    in_soa = False
    old_serial = None
    now = int(time.time())
    if 'serial_utc' in conf:
        serial = now
        serial_prefix = None
    else:
        serial = None
        serial_prefix = time.strftime('%Y%m%d', time.localtime(now))

    for line in oldfile.readlines():
        if not in_soa and line.startswith('@') and '(' in line:
            in_soa = True
        elif in_soa and not old_serial:
            # it is on this line!
            old_serial = line.split()[0]
            if serial_prefix and old_serial.startswith(serial_prefix):
                # we need to bump the number
                n = int(old_serial[len(serial_prefix):])
                serial = int(serial_prefix+'00') + n + 1
            elif serial_prefix:
                serial = serial_prefix+'00'
            print('\t\t\t{:>10}\t; Last changed by - {}'.format(serial,
                                                                user), file=newfile)
            continue
        print(line, end='', file=newfile)
    newfile.close()
    oldfile.close()
    conf['updated'].append(soa_new)

def myopts(o,a,conf):

    if o == '-u':
        conf['serial_utc'] = True

if __name__ == '__main__':

    conf = {
        'DEBUG': {},
        'updated': [],
        }

    conf,args = dnsutils.app_init('u', myopts, conf)
    for a in args:
        if a.endswith('.soa'):
            update_soa(a, conf)
    dnsutils.install_new(conf['updated'], conf)
    sys.exit(0)
